@@ -6,33 +6,104 @@ from abc import ABC, abstractmethod
 from dataclasses import dataclass
 from collections.abc import Iterable
 from enum import Enum
-from io import TextIOWrapper
+import io
 import itertools
 from pathlib import Path
-from debian.deb822 import Deb822, Packages, PkgRelation
+from typing import IO
+from debian.deb822 import Packages, PkgRelation
 from debian.debian_support import Version
 import logging
+import re
 from packageurl import PackageURL
 
+from ..util.checksum import ChecksumAlgo
 from .. import HAS_PYTHON_APT
 
 logger = logging.getLogger(__name__)
 
+# Debian policy 5.6.13, item 1
+DESC_REGEX_P1 = re.compile(r"^\s(\w.*)$")
 
-class ChecksumAlgo(Enum):
-    MD5SUM = 1
-    SHA1SUM = 2
-    SHA256SUM = 3
+
+class DpkgStatus(Enum):
+    NOT_INSTALLED = "n"
+    CONFIG_FILES = "c"
+    HALF_INSTALLED = "H"
+    UNPACKED = "U"
+    HALF_CONFIGURED = "F"
+    TRIGGERS_AWAITING = "W"
+    TRIGGERS_PENDING = "t"
+    INSTALLED = "i"
+    DEBSBOM_UNKNOWN = "0"
 
     @classmethod
-    def to_hashlib(cls, algo):
-        if algo == cls.MD5SUM:
-            return "md5"
-        if algo == cls.SHA1SUM:
-            return "sha1"
-        if algo == cls.SHA256SUM:
-            return "sha256"
-        raise NotImplementedError()
+    def from_dpkg(cls, status: str) -> "DpkgStatus":
+        if len(status) > 1:
+            status = status.lower()
+        for checkStatus in DpkgStatus:
+            if checkStatus == DpkgStatus.DEBSBOM_UNKNOWN:
+                continue
+            if status in [checkStatus.value, checkStatus.name.lower().replace("_", "-")]:
+                return checkStatus
+        raise ValueError(f"Unknown dpkg status '{status}'")
+
+
+class PkgListType(Enum):
+    """Type of package list data (e.g. PURL or dpkg status file)"""
+
+    STATUS_FILE = (0,)
+    PKG_LIST = (1,)
+    PURL_LIST = (2,)
+    ISAR_MANIFEST = (3,)
+
+
+class PkgListStream:
+    """
+    Wrapper around a packages iterator that takes care of closing the attached
+    stream (either on StopIteration or via context manager). It further allows
+    to return the source type of data from which the packages are created.
+    """
+
+    def __init__(self, stream: IO, kind: PkgListType, pkgs: Iterable["Package"]):
+        self._stream = stream
+        self.kind = kind
+        self._pkgs = pkgs
+
+    def __iter__(self) -> "PkgListStream":
+        return self
+
+    def __next__(self) -> "Package":
+        """return the next package, close the stream if no further elements."""
+        try:
+            return next(self._pkgs)
+        except StopIteration:
+            self.close()
+            raise
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        self.close()
+
+    def close(self):
+        """explicitly close the stream"""
+        self._stream.close()
+
+
+def filter_sources(pkgs: Iterable["Package"]) -> Iterable["SourcePackage"]:
+    """Free function to only return source packages"""
+    return filter(lambda p: p.is_source(), pkgs)
+
+
+def filter_binaries(pkgs: Iterable["Package"]) -> Iterable["BinaryPackage"]:
+    """Free function to only return binary packages"""
+    return filter(lambda p: p.is_binary(), pkgs)
+
+
+def filter_installed(pkgs: Iterable["BinaryPackage"]) -> Iterable["BinaryPackage"]:
+    """Free function to only return installed binary packages."""
+    return filter(lambda p: p.status in (DpkgStatus.INSTALLED, DpkgStatus.DEBSBOM_UNKNOWN), pkgs)
 
 
 @dataclass
@@ -78,40 +149,72 @@ class Package(ABC):
         self.version = Version(version)
 
     @classmethod
-    def parse_status_file(cls, status_file: Path) -> Iterable["Package"]:
+    def parse_status_file(cls, status_file: Path) -> PkgListStream:
         """
         Parse a dpkg status file and returns packages with their relations.
         """
         logger.info(f"Parsing status file '{status_file}'...")
-        return cls.inject_src_packages(cls._parse_status_file_raw(status_file))
+        deb822_stream = open(status_file, "r")
+        pkgs_it = cls.inject_src_packages(filter_installed(cls._parse_dpkg_status(deb822_stream)))
+        return PkgListStream(deb822_stream, PkgListType.STATUS_FILE, pkgs_it)
 
     @classmethod
-    def _parse_status_file_raw(cls, status_file: Path) -> Iterable["BinaryPackage"]:
+    def _parse_dpkg_status(
+        cls, stream: IO, force_no_apt: bool = False
+    ) -> Iterable["BinaryPackage"]:
         """
         Parse a dpkg status file and returns binary packages with their relations.
         The relations might contain links to packages that are not known yet (e.g.
         all source packages). These need to be resolved in a second pass.
+
+        The force_no_apt flag can be used to explicitly disable the use of python-apt.
+        Due to bugs in some python-apt versions, this might be required if a non-file
+        input stream is used.
         """
-        with open(status_file, "r") as status_file:
-            for package in Packages.iter_paragraphs(status_file, use_apt_pkg=HAS_PYTHON_APT):
-                bpkg = BinaryPackage.from_dep822(package)
-                logger.debug(f"Found binary package: '{bpkg.name}'")
-                yield bpkg
+        use_apt = HAS_PYTHON_APT and not force_no_apt
+        for package in Packages.iter_paragraphs(stream, use_apt_pkg=use_apt):
+            bpkg = BinaryPackage.from_deb822(package)
+            logger.debug(f"Found binary package: '{bpkg.name}'")
+            yield bpkg
 
     @classmethod
-    def parse_pkglist_stream(cls, stream: Iterable[str]) -> Iterable["Package"]:
+    def parse_pkglist_stream(cls, stream: IO) -> PkgListStream:
         """
         Parses a stream of space separated tuples describing packages
-        (name, version, arch) or PURLs alternatively. Each line describes one
-        package. Example:
+        (name, version, arch), PURLs, isar manifest data
+        (<src name>|<src version>|<bin name>:<bin arch>|<bin version>) or a
+        dpkg-status file. If not passing a dpkg-status file, each line describes
+        one package. Example:
         gcc 15.0-1 amd64
         g++ 15.0-1 amd64
         """
+        DPKG_STATUS_MAGIC = "Package: ".encode()
+        PURL_MAGIC = "pkg:deb/".encode()
+
+        if isinstance(stream, io.BufferedReader):
+            bstream = stream
+        elif isinstance(stream, io.TextIOBase):
+            bstream = io.BufferedReader(stream.buffer)
+        else:
+            bstream = io.BufferedReader(stream)
+
+        if bstream.peek(len(DPKG_STATUS_MAGIC)).startswith(DPKG_STATUS_MAGIC):
+            pkgs_it = cls.inject_src_packages(cls._parse_dpkg_status(bstream, force_no_apt=True))
+            return PkgListStream(bstream, PkgListType.STATUS_FILE, pkgs_it)
+        elif bstream.peek(len(PURL_MAGIC)).startswith(PURL_MAGIC):
+            return PkgListStream(
+                stream, PkgListType.PURL_LIST, map(lambda l: Package.from_purl(l.decode()), bstream)
+            )
+        elif sum(c == b"|"[0] for c in bstream.peek(128).split(b"\n")[0]) == 3:
+            return PkgListStream(
+                bstream, PkgListType.ISAR_MANIFEST, cls._parse_manifest_line_stream(bstream)
+            )
+        return PkgListStream(bstream, PkgListType.PKG_LIST, cls._parse_pkglist_line_stream(bstream))
+
+    @classmethod
+    def _parse_pkglist_line_stream(cls, stream: IO[bytes]) -> Iterable["Package"]:
         for line in stream:
-            if line.startswith("pkg:deb/"):
-                yield Package.from_purl(line)
-                continue
-            name, version, arch = line.split()
+            name, version, arch = line.decode().strip().split()
             if arch == "source":
                 yield SourcePackage(
                     name=name,
@@ -124,6 +227,26 @@ class Package(ABC):
                     version=version,
                 )
 
+    @classmethod
+    def _parse_manifest_line_stream(cls, stream: IO[bytes]) -> Iterable["Package"]:
+        """
+        Parse isar manifest file entries. The format is:
+        <src name>|<src version>|<bin name>:<bin arch>|<bin version>
+        Example:
+        json-c|0.16-2|libjson-c5:amd64|0.16-2
+        """
+        for line in stream:
+            src, srcv, bin, binv = line.decode().strip().split("|")
+            yield SourcePackage(name=src, version=srcv)
+            bin_parts = bin.split(":")
+            if len(bin_parts) != 2:
+                logger.error(f"Binary package '{bin}' misses the architecture specifier")
+            yield BinaryPackage(
+                name=bin_parts[0],
+                architecture=bin_parts[1] if len(bin_parts) == 2 else "unknown",
+                version=binv,
+            )
+
     @classmethod
     def from_purl(cls, purl: str) -> "Package":
         """
@@ -149,6 +272,15 @@ class Package(ABC):
             itertools.chain.from_iterable(map(lambda p: cls._resolve_sources(p, True), binpkgs))
         )
 
+    @classmethod
+    def referenced_src_packages(
+        cls, binpkgs: Iterable["BinaryPackage"]
+    ) -> Iterable["SourcePackage"]:
+        """Create and return referenced source packages"""
+        return cls._unique_everseen(
+            itertools.chain.from_iterable(map(lambda p: cls._resolve_sources(p, False), binpkgs))
+        )
+
     @classmethod
     def _unique_everseen(cls, iterable: Iterable[object], key=None):
         """
@@ -174,6 +306,7 @@ class Package(ABC):
             self.maintainer = other.maintainer
         if not self.homepage:
             self.homepage = other.homepage
+        self.checksums |= other.checksums
 
     @classmethod
     def _resolve_sources(cls, pkg: "BinaryPackage", add_pkg=False) -> Iterable["Package"]:
@@ -195,6 +328,18 @@ class Package(ABC):
         if add_pkg:
             yield pkg
 
+    def is_binary(self) -> bool:
+        """Helper to check if the package is a binary package"""
+        if isinstance(self, BinaryPackage):
+            return True
+        return False
+
+    def is_source(self) -> bool:
+        """Helper to check if the package is a source package"""
+        if isinstance(self, SourcePackage):
+            return True
+        return False
+
     @abstractmethod
     def purl(self) -> PackageURL:
         raise NotImplementedError
@@ -209,14 +354,38 @@ class Package(ABC):
         """Return the filename part from the locator of a package."""
         return self.locator.split("/")[-1]
 
+    def __str__(self) -> str:
+        return f"{self.name}@{self.version}"
+
+
+class VcsType(Enum):
+    Arch = "Arch"
+    Bzr = "Bazaar"
+    Cvs = "CVS"
+    Darcs = "Darcs"
+    Git = "Git"
+    Hg = "Mercurial"
+    Mtn = "Monotone"
+    Svn = "Subversion"
+
+
+@dataclass
+class VcsInfo:
+    """Internal representation of the Vcs-<type> information for a source package."""
+
+    type: VcsType
+    locator: str
+
 
 @dataclass(init=False)
 class SourcePackage(Package):
-    """Representation of a Debian Source package."""
+    """
+    Representation of a Debian Source package.
+    The checksums refer to the .dsc file.
+    """
 
     binaries: list[str] | None = None
-    vcs_browser: str | None = None
-    vcs_git: str | None = None
+    vcs: VcsInfo | None = None
     _locator: str | None = None
 
     def __init__(
@@ -226,8 +395,7 @@ class SourcePackage(Package):
         maintainer: str | None = None,
         binaries: list[str] = [],
         homepage: str | None = None,
-        vcs_browser: str | None = None,
-        vcs_git: str | None = None,
+        vcs: VcsInfo | None = None,
         checksums: dict[ChecksumAlgo, str] | None = None,
     ):
         self.name = name
@@ -235,8 +403,7 @@ class SourcePackage(Package):
         self.maintainer = maintainer
         self.binaries = binaries
         self.homepage = homepage
-        self.vcs_browser = vcs_browser
-        self.vcs_git = vcs_git
+        self.vcs = vcs
         self.checksums = checksums or {}
 
     def __hash__(self):
@@ -244,7 +411,7 @@ class SourcePackage(Package):
 
     def __eq__(self, other):
         # For compatibility reasons
-        if isinstance(other, SourcePackage):
+        if other.is_source():
             return self.purl() == other.purl()
         return NotImplemented
 
@@ -275,39 +442,57 @@ class SourcePackage(Package):
     def merge_with(self, other: "SourcePackage"):
         """Copy properties from other which are unset on our side. Merge lists."""
         super().merge_with(other)
-        if not self.vcs_browser:
-            self.vcs_browser = other.vcs_browser
-        if not self.vcs_git:
-            self.vcs_git = other.vcs_git
+        if not self.vcs:
+            self.vcs = other.vcs
         # add binaries from other
         binaries = list(self.binaries)
         binaries.extend(x for x in other.binaries if x not in binaries)
         self.binaries = binaries
 
     @staticmethod
-    def from_dep822(package) -> "SourcePackage":
+    def from_deb822(package) -> "SourcePackage":
         """
-        Create a package from a dep822 representation. If the dep822 input
+        Create a package from a deb822 representation. If the deb822 input
         is a .dsc file, the name is read from the source property.
         """
         name = package.get("Source") or package["Package"]
         version = Version(package.get("Version"))
         maintainer = package.get("Maintainer")
-        if package.get("Binaries") is not None:
-            binaries = [b.strip() for b in package["Binaries"].split(",")]
+        binary_entry = package.get("Binary")
+        if binary_entry is not None:
+            binaries = [b.strip() for b in binary_entry.split(",")]
         else:
             binaries = []
         homepage = package.get("Homepage")
-        vcs_browser = package.get("Vcs-Browser")
-        vcs_git = package.get("Vcs-Git")
+        vcs = None
+        for type in VcsType:
+            locator = package.get(f"Vcs-{type.name}")
+            if locator:
+                if vcs:
+                    logger.warning(
+                        f"Multiple VCS types found for package {name}: {vcs.type} and {type}"
+                    )
+                vcs = VcsInfo(type, locator)
+
+        # Checksums according to Debian policy 5.6.24
+        pkg_chksums = {}
+        for alg, deb822_field, chksm_name in [
+            (ChecksumAlgo.MD5SUM, "Files", "md5sum"),
+            (ChecksumAlgo.SHA1SUM, "Checksums-Sha1", "sha1"),
+            (ChecksumAlgo.SHA256SUM, "Checksums-Sha256", "sha256"),
+        ]:
+            chksums = package.get(deb822_field) or []
+            for c in filter(lambda c_: c_["name"].endswith(".dsc"), chksums):
+                pkg_chksums[alg] = c[chksm_name]
+
         return SourcePackage(
             name=name,
             version=version,
             maintainer=maintainer,
             binaries=binaries,
             homepage=homepage,
-            vcs_browser=vcs_browser,
-            vcs_git=vcs_git,
+            vcs=vcs,
+            checksums=pkg_chksums,
         )
 
 
@@ -322,6 +507,7 @@ class BinaryPackage(Package):
     built_using: list[Dependency]
     description: str | None
     manually_installed: bool
+    status: DpkgStatus
     _locator: str | None = None
 
     def __init__(
@@ -338,6 +524,7 @@ class BinaryPackage(Package):
         homepage: str | None = None,
         checksums: dict[ChecksumAlgo, str] | None = None,
         manually_installed: bool = True,
+        status: DpkgStatus = DpkgStatus.DEBSBOM_UNKNOWN,
     ):
         self.name = name
         self.section = section
@@ -351,12 +538,13 @@ class BinaryPackage(Package):
         self.homepage = homepage
         self.checksums = checksums or {}
         self.manually_installed = manually_installed
+        self.status = status
 
     def __hash__(self):
         return hash(self.purl())
 
     def __eq__(self, other):
-        if isinstance(other, BinaryPackage):
+        if other.is_binary():
             return self.purl() == other.purl()
         return NotImplemented
 
@@ -367,6 +555,22 @@ class BinaryPackage(Package):
             purl = purl + "?arch={}".format(self.architecture)
         return PackageURL.from_string(purl)
 
+    @property
+    def unique_depends(self):
+        """
+        Returns the unique dependencies without version.
+        The raw dependencies can include version specifiers, but as only a single
+        version can be installed at a time, we ignore them.
+        """
+        seen = set()
+        unique = []
+        for dep in self.depends:
+            key = (dep.name, dep.arch)
+            if key not in seen:
+                seen.add(key)
+                unique.append(dep)
+        return unique
+
     def merge_with(self, other: "BinaryPackage"):
         """Copy properties from other which are unset on our side. Merge lists and dicts. Or booleans."""
         super().merge_with(other)
@@ -378,8 +582,14 @@ class BinaryPackage(Package):
             self.source = other.source
         if not self.description:
             self.description = other.description
-        self.checksums |= other.checksums
         self.manually_installed |= other.manually_installed
+        # we cannot merge the status, but if the other package is
+        # marked as installed, consider all as installed.
+        if self.status == DpkgStatus.DEBSBOM_UNKNOWN:
+            self.status = other.status
+        elif other.status != DpkgStatus.DEBSBOM_UNKNOWN and self.status != other.status:
+            # this indicates an internal error
+            logger.warning(f"package statuses are inconsistent: {self.status} != {other.status}")
 
         depends = list(self.depends)
         depends.extend(x for x in other.depends if x not in depends)
@@ -406,9 +616,37 @@ class BinaryPackage(Package):
         self._locator = loc
 
     @staticmethod
-    def from_dep822(package) -> "BinaryPackage":
+    def _cleanup_description(desc: str | None) -> str | None:
         """
-        Create a ``BinaryPackage`` from a dep822 representation.
+        Parse description according to Debian policy 5.6.13
+        """
+        if desc is None:
+            return None
+        buffer = str()
+        in_paragraph = False
+        for line in desc.split("\n"):
+            matches = DESC_REGEX_P1.match(line)
+            if matches:
+                buffer += (" " if in_paragraph else "") + matches[1]
+                in_paragraph = True
+                continue
+
+            if len(line) == 0:
+                pass
+            elif line == " .":
+                buffer += "\n"
+            elif line[0] != " ":
+                # First line
+                buffer += line + "\n"
+            else:
+                buffer += "\n" + line[1:]
+            in_paragraph = False
+        return buffer.strip()
+
+    @classmethod
+    def from_deb822(cls, package) -> "BinaryPackage":
+        """
+        Create a ``BinaryPackage`` from a deb822 representation.
         """
         if package.source:
             srcdep = Dependency(package.source, None, ("=", package.source_version), arch="source")
@@ -423,15 +661,22 @@ class BinaryPackage(Package):
         sdepends = Dependency.from_pkg_relations(s_built_using, is_source=True)
 
         pkg_chksums = {}
-        for alg, dep822_name in [
+        for alg, deb822_name in [
             (ChecksumAlgo.MD5SUM, "MD5sum"),
             (ChecksumAlgo.SHA1SUM, "SHA1"),
             (ChecksumAlgo.SHA256SUM, "SHA256"),
         ]:
-            chksum = package.get(dep822_name)
+            chksum = package.get(deb822_name)
             if chksum:
                 pkg_chksums[alg] = chksum
 
+        status_raw = package.get("Status")
+        if status_raw:
+            _, _, pkg_status = status_raw.split(" ")
+            status = DpkgStatus.from_dpkg(pkg_status)
+        else:
+            status = DpkgStatus.DEBSBOM_UNKNOWN
+
         return BinaryPackage(
             name=package.get("Package"),
             section=package.get("Section"),
@@ -441,7 +686,8 @@ class BinaryPackage(Package):
             version=package.get("Version"),
             depends=dependencies,
             built_using=sdepends,
-            description=package.get("Description"),
+            description=cls._cleanup_description(package.get("Description")),
             homepage=package.get("Homepage"),
             checksums=pkg_chksums,
+            status=status,
         )
